# Licensed under the terms of http://www.apache.org/licenses/LICENSE-2.0
# Author (©): Alvaro del Castillo

from .scene import Scene
from .thing import Thing


class Block(Thing):

    def build(self):
        Scene.server.setBlock(self.position.x, self.position.y,
                              self.position.z, self.block)

        self._end_position = self.position



# Licensed under the terms of http://www.apache.org/licenses/LICENSE-2.0
# Author (©): Alvaro del Castillo

import mcpi.block
from mcpi.vec3 import Vec3

from .scene import Scene
from .thing import Thing


class Pyramid(Thing):

    height = 10

    def build(self):

        length = 2 * self.height - 1
        width = length

        for i in range(0, self.height):
            level = i
            Scene.server.setBlocks(
                self.position.x + level, self.position.y + level, self.position.z + level,
                self.position.x + (length - 1) - level,
                self.position.y + level,
                self.position.z + (width - 1) - level,
                self.block)

        self._end_position = Vec3(self.position.x + (length - 1),
                                  self.position.y + self.height - 1,
                                  self.position.z + (width - 1)
                                  )


class PyramidHollow(Thing):

    height = 10
    thick = 2

    def build(self):
        outer = Pyramid(self.position, self)
        outer.height = self.height
        outer.block = self.block
        outer.build()
        self._end_position = outer.end_position
        inner_x = self.position.x + self.thick
        inner_y = self.position.y
        inner_z = self.position.z + self.thick
        inner = Pyramid(Vec3(inner_x, inner_y, inner_z), self)
        inner.block = mcpi.block.AIR
        inner.height = self.height - self.thick
        inner.build()




# Licensed under the terms of http://www.apache.org/licenses/LICENSE-2.0
# Author (©): Alvaro del Castillo

from math import floor

from mcpi.vec3 import Vec3

from .scene import Scene
from .thing import Thing


class Bridge(Thing):

    large = 5
    height = None
    width = 1

    def build(self):
        for z in range(0, self.width):
            self.build_row(z)

        self._end_position = Vec3(self.position.x + self.large - 1,
                                  self.position.y + self.height,
                                  self.position.z + self.width - 1
                                  )

    def build_row(self, z):
        # large = 2 * height - 1
        max_height = floor((self.large + 1) / 2)
        if self.height is None:
            self.height = max_height

        for x in range(0, self.large):
            if x < max_height:
                y = x
            elif x == max_height and self.large % 2 == 0:
                y = max_height - 1
            else:
                y = y - 1

            final_y = y

            if self.height and y >= self.height-1:
                final_y = self.height-1

            Scene.server.setBlock(
                self.position.x + x, self.position.y + final_y, self.position.z + z,
                self.block)




# Licensed under the terms of http://www.apache.org/licenses/LICENSE-2.0
# Author (©): Alvaro del Castillo

import mcpi.block
import mcpi.vec3

from ._version import __version__

from .scene import Scene

from .utils import build_schematic_nbt


class Thing:
    """ base class for all objects in mcthings library """

    block = mcpi.block.BRICK_BLOCK
    """ block type used by the thing. Default to BRICK_BLOCK"""
    _block_empty = mcpi.block.AIR

    def __init__(self, position, parent=None):
        """
        Create a thing
        :param position: build position
        :param parent: parent Thing in which this one is included
        """

        self._end_position = None
        self._parent = parent
        self._position = None
        self._decorators = []

        if position:
            self._position = mcpi.vec3.Vec3(position.x, position.y, position.z)

        # Add then thing built to the scene
        if parent is None:
            Scene.add(self)

        # McThing version which created this Thing
        self._version = __version__

    @property
    def end_position(self):
        """ end position of the thing """
        return self._end_position

    @property
    def position(self):
        """ initial position of the thing """
        return self._position

    @property
    def parent(self):
        """ parent Thing in which this one is included """
        return self._position

    def build(self):
        """
        Build the thing and show it in Minecraft at position coordinates

        :return:
        """

    def unbuild(self):
        """
        Unbuild the thing in Minecraft

        :return:
        """

        block = self.block
        self.block = self._block_empty
        self.build()
        self.block = block

    def move(self, position):
        """
        Move the thing to a new position

        :param position: new position
        :return:
        """

        self.unbuild()
        self._position = position
        self.build()

    def to_schematic(self, file_path, blocks_data=False):
        """
        Convert the Thing to a Schematic Object

        :file_path: file in which to export the Thing in Schematic format
        :blocks_data: include blocks data (much slower)
        :return: the Schematic object
        """

        build_schematic_nbt(self.position, self.end_position, blocks_data).write_file(file_path)

    def add_decorator(self, decorator):
        """
        Add a new Decorator to be called once the Thing is decorated

        :param decorator: a Decorator to be called
        :return:
        """
        self._decorators.append(decorator)

    def decorate(self):
        """
        Call all decorators for the current Thing

        :return:
        """
        for decorator in self._decorators:
            decorator.decorate(self)



# TODO: at some point this must be a real Singleton

from mcpi.minecraft import Minecraft
# Licensed under the terms of http://www.apache.org/licenses/LICENSE-2.0
# Author (©): Alvaro del Castillo

from minecraftstuff import MinecraftDrawing

from mcthings.scene import Scene


class Decorator:
    """
    A Decorator is able to decorate a Thing based on its characteristics.

    If a Thing has decorators, they will be called after the build of the Thing.
    """

    @classmethod
    def decorate(cls, thing):
        """
        Decorate the thing

        :return:
        """
